//========================================================================================
//
// Car Finder Plugin
// 
//
// Paul Bennett
// bennett@cc.gatech.edu
//
//
//
//========================================================================================

#define START_X						10	// The X location of your fill point.
#define START_Y						65	// The Y location of your fill point.

#define EDGE_THRESHOLD_GREEN		10	// For flood filling (only used during texture fills) 
#define EDGE_THRESHOLD_BLUE			10
#define EDGE_THRESHOLD_RED			10
#define BINARY_THRESHOLD			5	// How many pixels must surround a pixel for it not to be noise
#define BLOB_SIZE_THRESHOLD			10	// How big should blobs be so they are not thrown out
#define PATTERN_MATCH_THRESHOLD_X	5	// How far in the X direction should we slide the pattern around.
#define PATTERN_MATCH_THRESHOLD_Y	5	// How far in the Y direction should we slide the pattern around.


#include "prTypes.h"
#include "prEffect.h"
#include "seedfill.h"
#include <stdio.h>
#include <math.h>

typedef struct {
	prPoint flood_point;
	long	last_frame;
} persistant_holder;

	char		welcomeMsg[255];
	long		testHwnd;

//========================================================================================
// Prototypes
//========================================================================================
PREMPLUGENTRY xFilter (short selector, VideoHandle theData);
void compute_delta(prPixel *frame1, prPixel *frame2, int *delta, long totalPix);
void edge_detection (int *delta, long totalPix, int nv);
void compute_center ( int * delta, prPoint *point, long width, long totalPix);
void get_trailing_edge ( int * delta, int *fillSpace, long totalPix);
void compute_vector (prPoint start, prPoint finish, prPoint *answer) ;
void get_leading_edge ( int * delta, int *fillSpace, long totalPix);
void noise_filter_binary (int *filter, long width, long totalPix, int *output) ;
void make_visible ( int *data, long totalPix)   ;
void number_blobs ( int *blob, long width, long height, long totalPix);
void remove_small_blobs ( int * blob, int *count, long totalPix);
void locate_north_pixel ( int *data, int *pixel_location, long totalPix );
void find_bounding_box ( int *data, long width, long totalPix, int *w, int *h);
void locate_north_pixel_single_frame ( int *data, long totalPix, int *pixel );
long best_fit_matcher ( int *bag, int *blobs, long nm_pixel, int *nm_blobs, int bag_width, int bag_height, long width, long height, long totalPix ) ;
void crosshair ( int *data, prPoint pt, unsigned char red, unsigned char green, unsigned char blue, long width, long totalPix) ;

//========================================================================================
// Perform the effect
//========================================================================================
PREMPLUGENTRY xFilter (short selector, VideoHandle theData)
{
	short		result = 0;
	long		height, width, rowbytes, totalPix;
	prRect		box;
	Ptr			srcpix, dstpix;

	Window		win;
	PPixHand	curPixHand, nextPixHand, twoPixHand, lastPixHand ;
	
	prPixel		*thisPixels, *nextPixels, *twoPixels, *lastPixels;
	persistant_holder	p_holder;
	long		i, j;
	int			*fillSpace12, *fillSpace1L; // holds the pixels that were flood filled
	int			*delta12, *delta1L	;		// holds the delta informtion
	int			*diff2, *diff3, *deltad2d3;  // delta of diff2 diff3
	int			*counting, *nm_pixels;
	int			*shadow12;
	prPoint		delta12_nm, delta23_nm, vector;
	short		err;
	int			pattern_width, pattern_height;
	int			shadow12_nm_pixel; // the north most pixel for shadow12
	int			shadow23_nm_pixel;


	testHwnd     = ((*theData)->piSuites->windFuncs->getMainWnd)();

	// Act according to the selector
	switch (selector)
	{
		case fsSetup:
			if ((*theData)->specsHandle)
			{
				// If we already had a specs handle, then we'd read our parameters out of it,
				// and put up a dialog with the parameters filled in
			}
			else
			{
				// Otherwise we'd put up a dialog with default values and allow the user to
				// change them. Then we'd allocate memory for a specs handle (using the memory
				// callback suite) in which to store the parameters.
			}
		
			((*theData)->piSuites->ppixFuncs->ppixGetBounds)((*theData)->source, &box);
			height = (long)box.bottom - (long)box.top;
			width = (long)box.right - (long)box.left;
	
			totalPix = height * width;

			rowbytes = ((*theData)->piSuites->ppixFuncs->ppixGetRowbytes)((*theData)->source);

			sprintf( welcomeMsg, "Welcome to Car Finder!.\n frame %d, total %d\nWindow Size: %d x %d (%d) -- %d", 
				 (*theData)->part, (*theData)->total, height, width, totalPix, rowbytes);
			//testHwnd = ((*theData)->piSuites->windFuncs->getMainWnd)();
			MessageBox( (HWND)testHwnd, welcomeMsg, &("Plugin Initializing"), 
						MB_OK);

			break;
		case fsExecute:
			((*theData)->piSuites->ppixFuncs->ppixGetBounds)((*theData)->destination, &box);
			height = (short)box.bottom - (short)box.top;
			width = (short)box.right - (short)box.left;
			rowbytes = ((*theData)->piSuites->ppixFuncs->ppixGetRowbytes)((*theData)->destination);
			totalPix = height * width;			
			srcpix = ((*theData)->piSuites->ppixFuncs->ppixGetPixels)((*theData)->source);
			dstpix = ((*theData)->piSuites->ppixFuncs->ppixGetPixels)((*theData)->destination);



		if ((*theData)->instanceData) {
			memcpy(&p_holder, *((*theData)->instanceData), sizeof(persistant_holder));
		} else {
			(*theData)->instanceData = (*theData)->piSuites->memFuncs->newHandleClear (sizeof(persistant_holder)); 

			// put in starting values.  when an interface is created, the stored info will be used here
			// to create these starting values
			p_holder.flood_point.x = START_X;  // fill in arbitrary spots now
			p_holder.flood_point.y = START_Y;	
			p_holder.last_frame = -1;

			memcpy(*((*theData)->instanceData), &p_holder, sizeof(persistant_holder));

		}

			// run this filter for each frame up to and including this frame since 
			// the filter was last run

// DEBUG
//sprintf( welcomeMsg, "last_frame: %d part %d total %d\n", p_holder.last_frame, (*theData)->part, (*theData)->total);
//MessageBox( (HWND)testHwnd, welcomeMsg, &("DEBUG"), MB_OK);


// DEBUG
//if ((*theData)->part == 0 && p_holder.last_frame == 0) assert(0);

			if ((((*theData)->part / (*theData)->tdb->sampleSize) + 2) > ((*theData)->total / (*theData)->tdb->sampleSize)) return result;

			for (i = (p_holder.last_frame + 1) ; i <= (((*theData)->part) /(*theData)->tdb->sampleSize) ; i++) {
				

// DEBUG
//sprintf( welcomeMsg, "last_frame: %d part %d i %d\n", p_holder.last_frame, (*theData)->part, i) ;
//MessageBox( (HWND)testHwnd, welcomeMsg, &("DEBUG"), MB_OK);
				

				// allocate memory for this blob
				fillSpace12 = malloc( totalPix * sizeof( int ));
				fillSpace1L= malloc( totalPix * sizeof( int ));
				delta12 = malloc( totalPix * sizeof( int ));
				delta1L = malloc( totalPix * sizeof( int ));
				diff2 = malloc( totalPix * sizeof( int ));
				diff3 = malloc( totalPix * sizeof( int ));
				deltad2d3 = malloc( totalPix * sizeof( int ));
				counting = malloc( totalPix * sizeof( int ));
				nm_pixels = malloc( totalPix * sizeof( int ));
				shadow12 = malloc( totalPix * sizeof( int ));

				
				// zero out the blob... maybe use calloc instead?
				for (j = 0; j < totalPix; j++) {
					fillSpace12[j] = false;
					fillSpace1L[j] = false;
					delta12[j] = false;
					delta1L[j] = false;
					diff2[j] = false;
					diff3[j] = false;
					deltad2d3[j] = false;
					counting[j] = false;
					nm_pixels[j] = false;
					shadow12[j] = false;

				}


				// we need to get a different frame than the current frame, so
				// Get the memory needed for the handle		
				curPixHand = ((*theData)->piSuites->ppixFuncs->ppixNew)(&box);
				// Get the frame
				err = ((*theData)->callBack)( ((*theData)->tdb->sampleSize)*i, curPixHand, &box, (*theData)->privateData );
				if (err) MessageBox( (HWND)testHwnd, &("error in callback 1"), &("Debug Info"), MB_OK);
				// Now get the frames pixels that we want to compare with the source frame
				thisPixels = (prPixel *) ((*theData)->piSuites->ppixFuncs->ppixGetPixels)(curPixHand);

				// get the next frame 
				nextPixHand = ((*theData)->piSuites->ppixFuncs->ppixNew)(&box);
				err = ((*theData)->callBack)( (*theData)->tdb->sampleSize*(i+1), nextPixHand, &box, (*theData)->privateData );
				if (err) MessageBox( (HWND)testHwnd, &("error in callback 2"), &("Debug Info"), MB_OK);
				nextPixels = (prPixel *) ((*theData)->piSuites->ppixFuncs->ppixGetPixels)(nextPixHand);

				// get the next frame plus one
				twoPixHand = ((*theData)->piSuites->ppixFuncs->ppixNew)(&box);
				err = ((*theData)->callBack)( (*theData)->tdb->sampleSize*(i+2), twoPixHand, &box, (*theData)->privateData );
				if (err) MessageBox( (HWND)testHwnd, &("error in callback 3"), &("Debug Info"), MB_OK);
				twoPixels = (prPixel *) ((*theData)->piSuites->ppixFuncs->ppixGetPixels)(twoPixHand);

				compute_delta(thisPixels, nextPixels, delta12, totalPix);
				// delta12 contains the difference between the current frame and curframe + 1

				// perform thresholding
				edge_detection(delta12, totalPix, true);

				// get the last frame plus one
				lastPixHand = ((*theData)->piSuites->ppixFuncs->ppixNew)(&box);
				err = ((*theData)->callBack)( (*theData)->total, lastPixHand, &box, (*theData)->privateData );
				if (err) MessageBox( (HWND)testHwnd, &("error in callback 3"), &("Debug Info"), MB_OK);
				lastPixels = (prPixel *) ((*theData)->piSuites->ppixFuncs->ppixGetPixels)(lastPixHand);

				// delta1L contains the difference between the current frame and last frame
				compute_delta(lastPixels, thisPixels, delta1L, totalPix);
				// perform thresholding
				edge_detection(delta1L, totalPix, true);

				// setup the bounding box
				win.x0 = 0;	win.x1 = width;	win.y0 = 0;	win.y1 = height;


				// fill the two binary shadows 
				// this picks out our edge vs. all the other edges in the frame
				binary_fill( p_holder.flood_point.x, p_holder.flood_point.y, &win, false , width,  delta12, fillSpace12);
				binary_fill( p_holder.flood_point.x, p_holder.flood_point.y, &win, false , width,  delta1L, fillSpace1L);

				// this is the output of the filter for the current frame.  
				// the rest of the code is preliminary processing for the next call to the filter.

				memcpy(dstpix, fillSpace1L, sizeof(prPixel)*height*width);	
				make_visible((int *) dstpix, totalPix);

				get_leading_edge(fillSpace12, fillSpace1L, totalPix);
				// fillSpace12 now contains the leading shadow w/ noise

				// get rid of some of the noise
				noise_filter_binary (fillSpace12, width, totalPix, shadow12);
				// shadow12 now contains the leading shadow

// View a noise cleaned shadow12
//make_visible(shadow12, totalPix);
//memcpy(dstpix, shadow12, sizeof(prPixel)*height*width);	
//return(result);

				compute_delta(lastPixels, nextPixels, diff2, totalPix);
				edge_detection(diff2, totalPix, true);

// If you want to view some of these steps
// in scrube mode, use this harness.  set i
// equal to the frame number
//				if (i == 4) {
//make_visible(diff2, totalPix);
//memcpy(dstpix, diff2, sizeof(prPixel)*height*width);	
//return(result);
//				}

				compute_delta(lastPixels, twoPixels, diff3, totalPix);
				edge_detection(diff3, totalPix, true);

				get_leading_edge(diff3, diff2, totalPix);
				// diff3 now contains the leading shadow23.

				// number all the blobs
				number_blobs ( diff3, width, height, totalPix) ;

				
				// remove smaller blobs
				remove_small_blobs ( diff3, counting, totalPix);


				// this locates all the north most pixels for all the blobs in diff3.
				locate_north_pixel ( diff3, nm_pixels, totalPix );


				// find the bounding box of Shadow12
				find_bounding_box ( shadow12 , width, totalPix, &pattern_width, &pattern_height); 
				
// Display the bounding box information
//sprintf( welcomeMsg, "width %d height %d\n", pattern_width, pattern_height);
//MessageBox( (HWND)testHwnd, welcomeMsg, &("NM Pixel"), MB_OK);
//return(result);

				
				locate_north_pixel_single_frame ( shadow12, totalPix, &shadow12_nm_pixel );

// Display the north most pixel
//sprintf( welcomeMsg, "%d", shadow12_nm_pixel);
//MessageBox( (HWND)testHwnd, welcomeMsg, &("NM Pixel"), MB_OK);


				// We chose to use the north most pixel for pattern matching.
				// If you want to use a different point, just pass those in like I passed in the
				// north most pixels.  The algorithm will not know the difference.
				shadow23_nm_pixel = best_fit_matcher ( shadow12, diff3, shadow12_nm_pixel, nm_pixels, pattern_width, pattern_height, width, height, totalPix );
					
				if (shadow23_nm_pixel == -1) {
// You can turn this messagebox on if you like, and it will tell you when it loses a shadow
// but that is annoying.
//				MessageBox( (HWND)testHwnd, &("Lost the shadows.  Try another point."), &("Best Fit Match"), MB_OK);

					vector.x = 0;
					vector.y = 0;
					
					
				} else {



					delta12_nm.x = (shadow12_nm_pixel % width);
					delta12_nm.y = (long) (shadow12_nm_pixel / width);

					delta23_nm.x = (shadow23_nm_pixel % width);
					delta23_nm.y = (long) (shadow23_nm_pixel / width);

					compute_vector(delta12_nm, delta23_nm, &vector);


				}

// Display some stats about the processing.
//sprintf( welcomeMsg, "orig %d %d\ndelta12 %d %d\ndelta23 %d %d\nvector %d %d\n new point: %d %d\n", p_holder.flood_point.x, p_holder.flood_point.y, delta12_nm.x, delta12_nm.y, delta23_nm.x, delta23_nm.y, vector.x, vector.y, p_holder.flood_point.x + vector.x, p_holder.flood_point.y + vector.y );
//MessageBox( (HWND)testHwnd, welcomeMsg, &("Debugging"), MB_OK);


// This will put the current point in the frame so we can see it.
// For debugging. 
//crosshair ( (int *) dstpix, p_holder.flood_point, 255, 0, 0, width, totalPix);

					p_holder.flood_point.x = p_holder.flood_point.x + vector.x;
					p_holder.flood_point.y = p_holder.flood_point.y + vector.y;


// This will put the new point in the frame so we can see it.
// For debugging. 
//crosshair ( (int *) dstpix, p_holder.flood_point, 0, 0, 255, width, totalPix);

				// clear the memory used
				((*theData)->piSuites->ppixFuncs->ppixDispose)(curPixHand);
				((*theData)->piSuites->ppixFuncs->ppixDispose)(nextPixHand);
				((*theData)->piSuites->ppixFuncs->ppixDispose)(twoPixHand);
				((*theData)->piSuites->ppixFuncs->ppixDispose)(lastPixHand);
				free(fillSpace12);
				free(fillSpace1L);
				free(delta12);
				free(delta1L);
				free(counting);
				free(nm_pixels);
				free(shadow12);
			
			}
	
			// let the future invocation know where we left of.  
			// basically, this is the frame in which the flood_point is valid.
			p_holder.last_frame = ((*theData)->part / ((*theData)->tdb->sampleSize));

			// copy any persistant info back so premiere can save it.
			memcpy(*((*theData)->instanceData), &p_holder, sizeof(persistant_holder));

			break;
		case fsDisposeData:  // time to release data in the InstanceData field;

			(*theData)->piSuites->memFuncs->disposeHandle((*theData)->instanceData);
		
			break;
	}
	return(result);
}


void compute_delta(prPixel *frame1, prPixel *frame2, int *delta, long totalPix) {

	long i;
	prPixel *temp;

	temp = (prPixel *) delta;

	for (i = 0; i < totalPix; i++) {
		temp[i].red = abs(frame2[i].red - frame1[i].red);
		temp[i].blue = abs(frame2[i].blue - frame1[i].blue);
		temp[i].green = abs(frame2[i].green - frame1[i].green);
	}

}


void edge_detection (int *delta, long totalPix, int nv) {

	long i;
	prPixel *temp;

	temp = (prPixel *) delta;

	for (i = 0; i < totalPix; i++) {
		if ((temp[i].red >= EDGE_THRESHOLD_RED) || 
		 (temp[i].blue >= EDGE_THRESHOLD_BLUE) ||
		 (temp[i].green >= EDGE_THRESHOLD_GREEN) )
		 delta[i] = nv; 
		else delta[i] = false;
	}

}



void get_leading_edge ( int * delta, int *fillSpace, long totalPix) {

	long i;
	

	for (i = 0; i < totalPix; i++) {
		if ( delta[i] != false) 
			if (fillSpace[i] != false)  // = for leading, != for trailing!
				delta[i] = false;
	}	

}


void get_trailing_edge ( int * delta, int *fillSpace, long totalPix) {

	long i;
	

	for (i = 0; i < totalPix; i++) {
		if ( delta[i] == true) 
			if (fillSpace[i] != true) 
				delta[i] = false;

	}	

}


// no longer used
//void compute_center ( int * delta, prPoint *point, long width, long totalPix) {
//
//	long i, x_total, y_total , num_points;
//
//	char		welcomeMsg[255];	
//
//	x_total = 0;
//	y_total = 0;
//	num_points = 0;
//
//	for (i = 0; i < totalPix; i++) {
//		if ( delta[i] == true) {
//			x_total =+ (i % width);
//			y_total =+ (long) (i / width);
//			num_points++;
//		}
//	}	
//
//
//	point->x = (long) x_total / num_points;
//	point->y = (long) y_total / num_points;
//
//
//
//}

void compute_vector (prPoint start, prPoint finish, prPoint *answer) {

answer->x = finish.x - start.x;
answer->y = finish.y - start.y;

}

void noise_filter_binary (int *filter, long width, long totalPix, int *output) {



long i, j;

for (i = 1; i < (totalPix / width) - 1; i++) {

	for (j = 1; j < width ; j++) {

		if ( (filter[((i-1)*width)+j-1] +
				filter[((i-1)*width)+j] +
				filter[((i-1)*width)+j+1] + 
				filter[(i*width)+j-1] +
				filter[(i*width)+j] +
				filter[(i*width)+j+1] +
				filter[((i+1)*width)+j-1] +
				filter[((i+1)*width)+j] +
				filter[((i+1)*width)+j+1]) >= BINARY_THRESHOLD)

			output[(i*width)+j] = true;
			else output[(i*width)+j] = false;

	}

}


}


void number_blobs ( int *blob, long width, long height, long totalPix) {


Window		win;
long		x, y, i, value;

value = 2;

win.x0 = 0;	win.x1 = width;	win.y0 = 0;	win.y1 = height;

for ( i = 0; i < totalPix; i++ ) {

	if (blob[i] == true) {
		x = (i % width);
		y = (long) (i / width);
		binary_fill( x, y, &win, value , width,  blob, NULL);
		value++;
	}


}

// Display the number of blobs numbered
//sprintf( welcomeMsg, "%d\n", value);
//MessageBox( (HWND)testHwnd, welcomeMsg, &("Number Blobs"), MB_OK);

}


void remove_small_blobs ( int * blob, int *count, long totalPix) {

long i, j, removed;

	removed = 0;

	for (i = 0; i < totalPix; i++ ) {

	count[blob[i]] = count[blob[i]] + 1;

	}

	for (i = 0; i < totalPix; i++) {

		if ((count[i] < BLOB_SIZE_THRESHOLD) && count[i] != 0) {

		// remove all of that color
		for ( j = 0; j < totalPix; j++)
			if (blob[j] == i) blob[j] = false;

			count[i] = 0; // since we removed this blob from the image, we need to zero it's counter
			removed++;
		}

	}

// Display the number of blobs removed
//sprintf( welcomeMsg, "%d", removed);
//MessageBox( (HWND)testHwnd, welcomeMsg, &("Number Blobs Removed"), MB_OK);


}



void locate_north_pixel ( int *data, int *pixel_location, long totalPix ) {

// since our bitmap is just one long array, the north most
// pixel is the first one you come upon.

	int n;

	for ( n = 0 ; n < totalPix; n++ ){
		if (data[n] != 0) {
			if (pixel_location[data[n]] == 0) pixel_location[data[n]] = n;

		}


	}


}

void locate_north_pixel_single_frame ( int *data, long totalPix, int *pixel ) {

// this will locate the north most pixel in a single frame.
// used to find the north most pixel of a frame with just one shadow in it.

	int n;

	*pixel = 0;

	for ( n = 0 ; n < totalPix; n++ )
		if (data[n] != 0) 
			if (*pixel == 0) *pixel = n;
}


void find_bounding_box ( int *data, long width, long totalPix, int *w, int *h) {
	// h and w are the height and width of the bounding box, OUT variables
	// width is the width of the clip

	long x, y, i, max_x, max_y, min_x, min_y;

	max_x = 0;
	max_y = 0;
	min_x = totalPix;
	min_y = totalPix;


		for ( i = 0; i < totalPix; i++) {
			if ( data[i] != 0 ) {
				x = (i % width);
				y = (long) (i / width);
				if ( x < min_x) min_x = x;
				if ( x > max_x) max_x = x;
				if ( y < min_y) min_y = y;
				if ( y > max_y) max_y = y;
			}

		}

		*w = max_x - min_x;
		*h = max_y - min_y;

}






void make_visible ( int *data, long totalPix)   {
// this is a debugging function.  make the binary pixels actualy show up.
	long i;
	

prPixel *pix;




	for (i = 0; i < totalPix; i++) 		
		if (data[i] != 0) {
			
		pix = (prPixel *) &(data[i]);

		pix->alpha = 255;
		pix->red = 0;
		pix->green = 255;
		pix->blue = 0;
	
		}
}


long best_fit_matcher ( int *bag, int *blobs, long nm_pixel, int *nm_blobs, int bag_width, int bag_height, long width, long height, long totalPix ) {


	long nm_x, nm_y, j, nm_blob_x, nm_blob_y, x ,y, dx, dy;
	long number_hits, max_hits_nm_pixel, max_hits;

	// max_hits_nm_pixel contains the nm pixel of the blob that had the most hits when maching.  this is
	// the nm pixel of our shadow!

	number_hits = 0;
	max_hits = 0;

	nm_x = (nm_pixel % width); // north most pixel of the bag
	nm_y = (long) (nm_pixel / width); // north most pixel of the bag

	max_hits_nm_pixel = -1;

	// compare this bounding box in the bag to each nm pixel in the array of blobs

	//assert(0);

	for (j = 0; j < totalPix; j++) {
		if (nm_blobs[j] != 0) {
			// we found a nm pixel in our list.
			
			nm_blob_x = (nm_blobs[j] % width);
			nm_blob_y = (long) (nm_blobs[j] / width);

			for ( dx = (0 - PATTERN_MATCH_THRESHOLD_X) ; dx <= PATTERN_MATCH_THRESHOLD_X	; dx++) {	
				for ( dy = (0 - PATTERN_MATCH_THRESHOLD_Y) ; dy <= PATTERN_MATCH_THRESHOLD_Y	; dy++) {	
					number_hits = 0;
					for (x = 0; x < bag_width; x++ ) {
						for ( y = 0; y < bag_height; y++ ) {
							if ((nm_y+y <= height) && (nm_x+x <= width) && // checking bounds
								(nm_blob_y+y+dy <= height) && (nm_blob_x+x+dx <= width)) {
								if ((bag[((nm_y+y)*width)+(nm_x+x)] != 0) && (blobs[((nm_blob_y+y+dy)*width)+(nm_blob_x+x+dx)] != 0))  {
									number_hits++;
									if (number_hits > max_hits) {
										max_hits = number_hits;
										max_hits_nm_pixel = ((nm_blob_y+dy)*width)+(nm_blob_x+dx);
									}
								}
							}
						}
					}
				}
			}
		}
	}

	return max_hits_nm_pixel;

}



void crosshair ( int *data, prPoint pt, unsigned char red, unsigned char green, unsigned char blue, long width, long totalPix) {


prPixel *pix;

pix = (prPixel *) &(data[(pt.y*width)+pt.x]);

pix->red = red;
pix->green = green;
pix->blue = blue;

pix = (prPixel *) &(data[(pt.y*width)+pt.x+1]);

pix->red = red;
pix->green = green;
pix->blue = blue;

pix = (prPixel *) &(data[(pt.y*width)+pt.x-1]);

pix->red = red;
pix->green = green;
pix->blue = blue;

pix = (prPixel *) &(data[((pt.y+1)*width)+pt.x]);

pix->red = red;
pix->green = green;
pix->blue = blue;

pix = (prPixel *) &(data[((pt.y-1)*width)+pt.x]);

pix->red = red;
pix->green = green;
pix->blue = blue;



}