/*
 * A Seed Fill Algorithm
 * by Paul Heckbert
 * from "Graphics Gems", Academic Press, 1990
 *
 * user provides pixelread() and pixelwrite() routines
 */

/*
 * fill.c : simple seed fill program
 * Calls pixelread() to read pixels, pixelwrite() to write pixels.
 *
 * Paul Heckbert	13 Sept 1982, 28 Jan 1987
 */

#include "seedfill.h"
#include "prTypes.h"
#include <math.h>

/*
 * fill: set the pixel at (x,y) and all of its 4-connected neighbors
 * with the same pixel value to the new pixel value nv.
 * A 4-connected neighbor is a pixel above, below, left, or right of a pixel.
 */

int close_enuf( Pixel pone, Pixel ptwo ) ;

long fill(x, y, win, nv, width, blobArray, fillSpace)
int x, y;	    /* seed point      */
Window *win;	/* screen window   */
Pixel nv;	    /* new pixel value */
int width;      /* width of screen */
int *blobArray; /* blobArray       */
int *fillSpace; /* image showing where the fill was */

{
    int l, x1, x2, dy;
	long retVal = 0;
    Pixel ov;	/* old pixel value */
    Segment stack[MAX], *sp = stack;	/* stack of filled segments */

    ov = pixelread(x, y, width, blobArray);		/* read pv at seed point */
    if (close_enuf(ov, nv) || x<win->x0 || x>win->x1 || y<win->y0 || y>win->y1) return( 0 );

	//MessageBox( (HWND) parWnd, "Getting past this", &(" yada" ), MB_OK );
    PUSH(y, x, x, 1);			/* needed in some cases */
    PUSH(y+1, x, x, -1);		/* seed segment (popped 1st) */

    while (sp>stack) {
	/* pop segment off stack and fill a neighboring scan line */
	POP(y, x1, x2, dy);
	/*
	 * segment of scan line y-dy for x1<=x<=x2 was previously filled,
	 * now explore adjacent pixels in scan line y
	 */
	for (x=x1; x>=win->x0 && close_enuf(pixelread(x, y, width, blobArray), ov); x--) {
	    retVal += pixelwrite(x, y, nv, width, blobArray, fillSpace);
	
	}
	if (x>=x1) goto skip;
	l = x+1;
	if (l<x1) PUSH(y, l, x1-1, -dy);		/* leak on left? */
	x = x1+1;
	do {
		for (; x<=win->x1 && close_enuf(pixelread(x, y, width, blobArray), ov); x++) {
			retVal += pixelwrite(x, y, nv, width, blobArray, fillSpace);
			
		}
		PUSH(y, l, x-1, dy);
	    if (x>x2+1) PUSH(y, x2+1, x-1, -dy);	/* leak on right? */
skip:	    for (x++; x<=x2 && (! close_enuf(pixelread(x, y, width, blobArray), ov)); x++);
		l = x;
	} while (x<=x2);
    }

	return( retVal );
}


int pixelread( int x, int y, int width, int *blobArray ) {
	
	long pixelLocation;

	pixelLocation = ( width * y ) + x;
	
	return( blobArray[ pixelLocation ] );

}

long pixelwrite( int x, int y, int nv, int width, int *blobArray, int *fillSpace) {

	long pixelLocation;

	pixelLocation = ( width * y ) + x;

	blobArray[ pixelLocation ] = nv;

	if (fillSpace != NULL) fillSpace[ pixelLocation ] = true;

	return( pixelLocation );

}


int close_enuf( Pixel pone, Pixel ptwo ) {

prPixel* pix_one;
prPixel* pix_two;

pix_one = (prPixel *) &pone;
pix_two = (prPixel *) &ptwo;

if ((abs(pix_one->red - pix_two->red) < 40) &&
	(abs(pix_one->green - pix_two->green) < 40) &&
	(abs(pix_one->blue - pix_two->blue) < 40)) 
	return true;

return false;
}





long binary_fill(x, y, win, nv, width, blobArray, fillSpace)
int x, y;	    /* seed point      */
Window *win;	/* screen window   */
Pixel nv;	    /* new pixel value */
int width;      /* width of screen */
int *blobArray; /* blobArray       */
int *fillSpace; /* blob info     */
{
    int l, x1, x2, dy;
	long retVal = 0;
    Pixel ov;	/* old pixel value */
    Segment stack[MAX], *sp = stack;	/* stack of filled segments */

    ov = pixelread(x, y, width, blobArray);		/* read pv at seed point */
    if (ov==nv || x<win->x0 || x>win->x1 || y<win->y0 || y>win->y1) return( 0 );

	//MessageBox( (HWND) parWnd, "Getting past this", &(" yada" ), MB_OK );
    PUSH(y, x, x, 1);			/* needed in some cases */
    PUSH(y+1, x, x, -1);		/* seed segment (popped 1st) */

    while (sp>stack) {
	/* pop segment off stack and fill a neighboring scan line */
	POP(y, x1, x2, dy);
	/*
	 * segment of scan line y-dy for x1<=x<=x2 was previously filled,
	 * now explore adjacent pixels in scan line y
	 */
	for (x=x1; x>=win->x0 && pixelread(x, y, width, blobArray)==ov; x--) {
	    retVal += pixelwrite(x, y, nv, width, blobArray, fillSpace);
	}
	if (x>=x1) goto skip;
	l = x+1;
	if (l<x1) PUSH(y, l, x1-1, -dy);		/* leak on left? */
	x = x1+1;
	do {
		for (; x<=win->x1 && pixelread(x, y, width, blobArray)==ov; x++) {
			retVal += pixelwrite(x, y, nv, width, blobArray, fillSpace);
		}
		PUSH(y, l, x-1, dy);
	    if (x>x2+1) PUSH(y, x2+1, x-1, -dy);	/* leak on right? */
skip:	    for (x++; x<=x2 && pixelread(x, y, width, blobArray)!=ov; x++);
		l = x;
	} while (x<=x2);
    }

	return( retVal );
}


/*
int pixelread( int x, int y ) { };
int pixelwrite( int x, int y, int nv ) { }; 
*/